'''

@author: rch
'''


from traits.api import \
    HasStrictTraits, Instance, Button, Event, \
    DelegatesTo
from traits.etsconfig.api import ETSConfig
from traitsui.api import \
    TreeEditor, TreeNode, View, Item, VGroup, \
    ToolBar, \
    HSplit
from traitsui.menu import \
    Menu, MenuBar, Separator
from ibvpy.view.ui.bmcs_tree_node import \
    BMCSRootNode, BMCSTreeNode, BMCSLeafNode

import traits.api as tr

from .bmcs_study import BMCSStudy
# from .bmcs_tree_view_handler import \
#     menu_tools_report_pdf, menu_tools_report_tex, \
#     BMCSTreeViewHandler, plot_self, menu_save, \
#     menu_open, menu_exit, \
#     toolbar_actions, key_bindings
from .i_bmcs_model import IBMCSModel


if ETSConfig.toolkit == 'wx':
    from traitsui.wx.tree_editor import \
        DeleteAction
if ETSConfig.toolkit == 'qt4':
    from traitsui.qt4.tree_editor import \
        DeleteAction
else:
    raise ImportError("tree actions for %s toolkit not available" %
                      ETSConfig.toolkit)


# tree_node = TreeNode(node_for=[BMCSRootNode, BMCSTreeNode],
#                      auto_open=False,
#                      children='tree_node_list',
#                      label='node_name',
#                      view='tree_view',
#                      menu=Menu(plot_self, DeleteAction),
#                      )
#
# leaf_node = TreeNode(node_for=[BMCSLeafNode],
#                      auto_open=True,
#                      children='',
#                      label='node_name',
#                      view='tree_view',
#                      menu=Menu(plot_self)
#                      )
#
# tree_editor = TreeEditor(
#     nodes=[tree_node, leaf_node],
#     selected='selected_node',
#     orientation='vertical'
# )


class BMCSWindow(BMCSStudy):

    selected_node = Instance(HasStrictTraits)

    def _selected_node_changed(self):
        self.selected_node.ui = self

    def get_vot_range(self):
        return self.viz_sheet.get_vot_range()

    vot = DelegatesTo('viz_sheet')

    data_changed = Event

    replot = Button

    def _replot_fired(self):
        self.figure.clear()
        self.selected_node.plot(self.figure)
        self.data_changed = True

    clear = Button()

    def _clear_fired(self):
        self.figure.clear()
        self.data_changed = True

    view = View(
        HSplit(
            VGroup(
                Item('model',
                     id='bmcs.hsplit.left.tree.id',
                     dock='tab',
                     # editor=tree_editor,
                     resizable=True,
                     show_label=False,
                     width=300,
                     height=200,
                     ),
                #                Item('selected_node@'),
                id='bmcs.hsplit.left.id',
                dock='tab',
            ),
            VGroup(
                Item('viz_sheet@',
                     show_label=False,
                     id='bmcs.hsplit.viz3d.notebook.id',
                     dock='tab',
                     ),
                dock='tab',
                id='bmcs.hsplit.viz3d.id',
                label='viz sheet',
            ),
            dock='tab',
            id='bmcs.hsplit.id',
        ),
        #        dock='tab',
        id='bmcs.id',
        width=1.0,
        height=1.0,
        title='BMCS',
        resizable=True,
        # handler=BMCSTreeViewHandler(),
        # key_bindings=key_bindings,
        # toolbar=ToolBar(*toolbar_actions,
        #                 image_size=(32, 32),
        #                 show_tool_names=False,
        #                 show_divider=True,
        #                 name='view_toolbar'),
        # menubar=MenuBar(Menu(menu_exit, Separator(),
        #                      menu_save, menu_open,
        #                      name='File'),
        #                 Menu(menu_tools_report_tex,
        #                      menu_tools_report_pdf,
        #                      name='Tools'),
        #
        #                 )
    )


#if __name__ == '__main__':
    from ibvpy.view.examples.response_tracer import ResponseTracer
    from ibvpy.sim.sim_base import Simulator
    # from ibvpy.core.bcond_mngr import BCondMngr
    from ibvpy.bcond import BCDof, BCSlice
    # bc_mngr = BCondMngr()
    # bc_mngr.bcond_list = [
    #     BCDof(),
    #     BCSlice()]
    # rt = ResponseTracer()
    # bm = Simulator(node_name='sim',
    #                tree_node_list=[BMCSTreeNode(node_name='subnode 1'),
    #                                BMCSTreeNode(node_name='subnode 2'),
    #                                rt,
    #                                bc_mngr
    #                                ])
    # bm.tree_node_list += [bm.tline]
    # tv = BMCSWindow(model=bm)
    # rt.add_viz2d('time_profile', 'time profile')
    # tv.configure_traits()




from math import pi as Pi, cos, sin, exp, sqrt as scalar_sqrt

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from ibvpy.tmodel.mats_eval import IMATSEval
from numpy import \
    array, zeros, dot, \
    sqrt, vdot, \
    float_, diag
from scipy.linalg import eig, inv, norm
from traits.api import \
    Array, Bool, Callable, Enum, Float, HasTraits, \
    Instance, Int, Trait, Range, HasTraits, on_trait_change, Event, \
    Dict, Property, cached_property, Delegate
from traitsui.api import \
    Item, View, HSplit, VSplit, VGroup, Group, Spring
from util.traits.either_type import EitherType

import numpy as np

from .yield_face2D import IYieldFace2D, J2, DruckerPrager, Gurson, CamClay


#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS2DPlastic(MATS2DEval):
    '''
    Elastic Model.
    '''

    # implements(IMATSEval)

    #-------------------------------------------------------------------------
    # Parameters of the numerical algorithm (integration)
    #-------------------------------------------------------------------------

    stress_state = Enum("plane strain", "plane stress",)
    algorithm = Enum("closest point", "cutting plane")

    #-------------------------------------------------------------------------
    # Material parameters
    #-------------------------------------------------------------------------
    yf = EitherType(klasses=[J2, DruckerPrager, Gurson, CamClay],
                    label="Yield Face",
                    desc="Yield Face Definition"
                    )
#    yf = Instance( IYieldFace2D,
#                 label = "Yield Face",
#                 desc = "Yield Face Definition")
    E = Float(210.0e+3,
              label="E",
              desc="Young's Modulus")
    nu = Float(0.2,
               label='nu',
               desc="Poison's ratio")
    K_bar = Float(0.,
                  label='K',
                  desc="isotropic softening parameter")
    H_bar = Float(0.,
                  label='H',
                  desc="kinematic softening parameter")
    tolerance = Float(1.0e-4,
                      label='TOL',
                      desc="tolerance of return mapping")

    max_iter = Int(20,
                   label='Iterations',
                   desc="maximal number of iterations")

    D_el = Property(Array(float), depends_on='E, nu')

    @cached_property
    def _get_D_el(self):
        if self.stress_state == "plane_stress":
            return self._get_D_plane_stress()
        else:
            return self._get_D_plane_strain()

    H_mtx = Property(Array(float), depends_on='K_bar, H_bar')

    @cached_property
    def _get_H_mtx(self):
        H_mtx = diag([self.K_bar, self.H_bar, self.H_bar, self.H_bar])
        return H_mtx

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #

    #-------------------------------------------------------------------------
    # View specification
    #-------------------------------------------------------------------------

    view_traits = View(VSplit(Group(Item('yf'),
                                    Item('E'),
                                    Item('nu'),
                                    Item('K_bar'),
                                    Item('H_bar'),
                                    Item('tolerance'),
                                    Item('max_iter')),
                              Group(Item('stress_state', style='custom'),
                                    Item('algorithm', style='custom'),
                                    Spring(resizable=True),
                                    label='Configuration parameters', show_border=True,
                                    ),
                              ),
                       resizable=True
                       )

    #-------------------------------------------------------------------------
    # Private initialization methods
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------
    def get_state_array_size(self):
        '''
        Return number of number to be stored in state array
        @param sctx:spatial context
        '''
        return 7

    def setup(self, sctx):
        '''
        Intialize state variables.
        '''
        sctx.mats_state_array = zeros(7, float_)

    def new_cntl_var(self):
        return zeros(3, float_)

    def new_resp_var(self):
        return zeros(3, float_)

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        delta_gamma = 0.
        if sctx.update_state_on:
            # print "in us"
            eps_n = eps_app_eng - d_eps
            sigma, f_trial, epsilon_p, q_1, q_2 = self._get_state_variables(
                sctx, eps_n)

            sctx.mats_state_array[:3] = epsilon_p
            sctx.mats_state_array[3] = q_1
            sctx.mats_state_array[4:] = q_2

        diff1s = zeros([3])
        sigma, f_trial, epsilon_p, q_1, q_2 = self._get_state_variables(
            sctx, eps_app_eng)
        # Note: the state variables are not needed here, just gamma
        diff2ss = self.yf.get_diff2ss(eps_app_eng, self.E, self.nu, sctx)
        Xi_mtx = inv(inv(self.D_el) + delta_gamma * diff2ss * f_trial)
        N_mtx_denom = sqrt(dot(dot(diff1s, Xi_mtx), diff1s))
        if N_mtx_denom == 0.:
            N_mtx = zeros(3)
        else:
            N_mtx = dot(Xi_mtx, self.diff1s) / N_mtx_denom
        D_mtx = Xi_mtx - vdot(N_mtx, N_mtx)

        # print "sigma ",sigma
        # print "D_mtx ",D_mtx
        return sigma, D_mtx

    #-------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #-------------------------------------------------------------------------
    def _get_state_variables(self, sctx, eps_app_eng):
        epsilon_p = eps_p_old = sctx.mats_state_array[:3]
        q_1 = sctx.mats_state_array[3]
        q_2 = sctx.mats_state_array[4:]

        sigma = dot(self.D_el, (eps_app_eng - epsilon_p))
        xi_trial = sigma - q_2
        f_trial = self.yf.get_f_trial(xi_trial, q_1)
        # print "f_trial", f_trial
        R_k = zeros(3)
        int_count = 1
        while f_trial > self.tolerance or norm(R_k) > self.tolerance:
            if int_count > self.max_iter:
                print("Maximal number of iteration reached")
                break
            diff1s = self.yf.get_diff1s(
                eps_app_eng, self.E, self.nu, sctx)
            diff1q = self.yf.get_diff1q(eps_app_eng, self.E, self.nu, sctx)

            if self.stress_state == "plane stress":
                raise NotImplementedError
            else:
                if self.algorithm == "cutting plane":
                    delta_gamma_2 = (f_trial / 
                                     (dot(dot(diff1s, self.D_el), diff1s) + 
                                      dot(dot(diff1q, self.H_mtx), diff1q)))

                    delta_gamma += delta_gamma_2
                    epsilon_p += delta_gamma_2 * diff1s

                    q_1 += delta_gamma_2 * self.K_bar * diff1q[0]
                    q_2 += delta_gamma_2 * self.H_bar * diff1q[1:]

                elif self.algorithm == "closest point":
                    diff2ss = self.yf.get_diff2ss(
                        eps_app_eng, self.E, self.nu, sctx)
                    diff2sq = self.yf.get_diff2sq(
                        eps_app_eng, self.E, self.nu, sctx)
                    A_mtx_inv = inv(self.D_el) + delta_gamma * diff2ss
                    # self.delta_gamma *  diff2sq))
                    A_mtx = inv(A_mtx_inv)
                    delta_gamma_2 = ((f_trial - dot(dot(diff1s, A_mtx), R_k)) / 
                                     (dot(dot(diff1s, A_mtx), diff1s)))

                    delta_eps_p = dot(dot(inv(self.D_el), A_mtx),
                                      R_k + delta_gamma_2 * diff1s)

                    epsilon_p += delta_eps_p
                    delta_gamma += delta_gamma_2

                    R_k = -epsilon_p + eps_p_old + delta_gamma * diff1s

                else:
                    raise NotImplementedError

            sigma = dot(self.D_el, (eps_app_eng - epsilon_p))
            xi_trial = sigma - q_2

            f_trial = self.yf.get_f_trial(xi_trial, q_1)
            # print "f_trial_after", self.f_trial
            int_count += 1
        return sigma, f_trial, epsilon_p, q_1, q_2

    def _get_D_plane_stress(self):
        E = self.E
        nu = self.nu
        D_stress = zeros([3, 3])
        D_stress[0][0] = E / (1.0 - nu * nu)
        D_stress[0][1] = E / (1.0 - nu * nu) * nu
        D_stress[1][0] = E / (1.0 - nu * nu) * nu
        D_stress[1][1] = E / (1.0 - nu * nu)
        D_stress[2][2] = E / (1.0 - nu * nu) * (1.0 / 2.0 - nu / 2.0)
        return D_stress

    def _get_D_plane_strain(self):
        E = self.E
        nu = self.nu
        D_strain = zeros([3, 3])
        D_strain[0][0] = E * (1.0 - nu) / (1.0 + nu) / (1.0 - 2.0 * nu)
        D_strain[0][1] = E / (1.0 + nu) / (1.0 - 2.0 * nu) * nu
        D_strain[1][0] = E / (1.0 + nu) / (1.0 - 2.0 * nu) * nu
        D_strain[1][1] = E * (1.0 - nu) / (1.0 + nu) / (1.0 - 2.0 * nu)
        D_strain[2][2] = E * (1.0 - nu) / (1.0 + nu) / (2.0 - 2.0 * nu)
        return D_strain

    #-------------------------------------------------------------------------
    # Response trace evaluators
    #-------------------------------------------------------------------------

    def get_sig_norm(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return array([scalar_sqrt(sig_eng[0] ** 2 + sig_eng[1] ** 2)])

    def get_eps_p(self, sctx, eps_app_eng):
        # print "eps tracer ", sctx.mats_state_array[:3]
        return sctx.mats_state_array[:3]

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app': self.get_sig_app,
                'eps_app': self.get_eps_app,
                'sig_norm': self.get_sig_norm,
                'eps_p': self.get_eps_p}


if __name__ == '__main__':
    #-------------------------------------------------------------------------
    # Example using the mats2d_explore
    #-------------------------------------------------------------------------
    from ibvpy.api import RTDofGraph
    from ibvpy.tmodel.mats2D.mats2D_explore import MATS2DExplore
    from .yield_face2D import J2
    mats2D_explore = \
        MATS2DExplore(mats2D_eval=MATS2DPlastic(yf=J2()),
                      rtrace_list=[RTDofGraph(name='strain 0 - stress 0',
                                               var_x='eps_app', idx_x=0,
                                               var_y='sig_app', idx_y=0,
                                               record_on='update'),
                                   RTDofGraph(name='strain 0 - strain 1',
                                               var_x='eps_app', idx_x=0,
                                               var_y='eps_app', idx_y=1,
                                               record_on='update'),
                                   RTDofGraph(name='stress 0 - stress 1',
                                               var_x='sig_app', idx_x=0,
                                               var_y='sig_app', idx_y=1,
                                               record_on='update'),
                                   RTDofGraph(name='time - sig_norm',
                                               var_x='time', idx_x=0,
                                               var_y='sig_norm', idx_y=0,
                                               record_on='update')

                                   ])

    mats2D_explore.tloop.eval()
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp(ibv_resource=mats2D_explore)
    ibvpy_app.main()
#    from ibvpy.core.sdomain import SDomain
#    mm = MATS2DPlastic(E = 1.,
#                       nu = 0.,
#                       yf = J2(sigma_y = 1.))
#
#    eps_app_eng = array([1.5,0.,0.])
#    d_eps = array([1.,0.,0.])
#    sctx = SDomain()
#    sctx.update_state_on = False
#    sctx.mats_state_array = zeros(7)
#    tn = tn1 = 0.
#
#    sig, D = mm.get_corr_pred(sctx, eps_app_eng, d_eps, tn, tn1)
#
#    print sig



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 8, 2009 by: rch
from ibvpy.tmodel.mats1D import \
    MATS1DElastic, MATS1DPlastic, MATS1DDamage
from ibvpy.tmodel.mats1D5.mats1D5_eval import MATS1D5Eval
from ibvpy.tmodel.mats_eval import IMATSEval
from numpy import zeros, zeros_like, array
from traits.api import provides, \
    Instance, Property, cached_property, List, \
    Callable, String, Int, HasTraits
from traitsui.api import \
    View, Item
from ibvpy.util.traits.either_type import \
    EitherType

mats_klasses = [MATS1DElastic,
                MATS1DPlastic,
                MATS1DDamage]


@provides(IMATSEval)
class MATS1D5Bond(MATS1D5Eval):

    '''Bond model for two phases interacting over an interface with zero thickness.

    Both phases can be associated with arbitrary 1D tmodel model.
    Interface behavior can be defined for both sliding and opening using 1D tmodel models.
    '''

    #-------------------------------------------------------------------------
    # Submodels constituting the interface behavior
    #-------------------------------------------------------------------------

    mats_phase1 = EitherType(klasses=mats_klasses,
                             desc='Material model of phase 1')
    mats_ifslip = EitherType(klasses=mats_klasses,
                             desc='Material model for interface slippage')
    mats_ifopen = EitherType(klasses=mats_klasses,
                             desc='Material model for interface opening')
    mats_phase2 = EitherType(klasses=mats_klasses,
                             desc='Material model for phase 2')

    def _mats_phase1_default(self):
        return MATS1DElastic()

    def _mats_ifslip_default(self):
        return MATS1DElastic()

    def _mats_ifopen_default(self):
        return MATS1DElastic()

    def _mats_phase2_default(self):
        return MATS1DElastic()

    traits_view = View(Item('mats_phase1@'),
                       Item('mats_ifslip@'),
                       Item('mats_ifopen@'),
                       Item('mats_phase2@'),
                       resizable=True,
                       scrollable=True,
                       width=0.8,
                       height=0.9,
                       buttons=['OK', 'Cancel'])

    #-------------------------------------------------------------------------
    # Subsidiary maps to enable generic loop over the material models
    #-------------------------------------------------------------------------

    # order of the material model names used for association in generic loops
    _mats_names = List(
        ['mats_phase1', 'mats_slip', 'mats_open', 'mats_phase2'])

    # hidden list of all four involved tmodel
    _mats_list = Property(
        depends_on='mats_phase1, mats_slip, mats_open, mats_phase2')

    @cached_property
    def _get__mats_list(self):
        return [self.mats_phase1, self.mats_ifslip, self.mats_ifopen, self.mats_phase2]

    # state array map with the sizes of the array
    _state_sizes = Property(
        depends_on='mats_phase1, mats_slip, mats_open, mats_phase2')

    @cached_property
    def _get__state_sizes(self):
        return array([mats.get_state_array_size()
                      for mats in self._mats_list], dtype='int_')

    # state array map with the sizes of the array
    _state_offsets = Property(
        depends_on='mats_phase1, mats_slip, mats_open, mats_phase2')

    @cached_property
    def _get__state_offsets(self):
        offsets = zeros_like(self._state_sizes)
        offsets[1:] = self._state_sizes[:-1].cumsum()
        return offsets

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------

    def get_state_array_size(self):
        '''
        Return the number of floats to be saved
        '''
        return self._state_sizes.cumsum()[-1]

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps_app_eng, tn, tn1, *args, **kw):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''

        sig_app_eng = zeros_like(eps_app_eng)
        # @todo [rch] dirty - when called form response tracer
        if isinstance(d_eps_app_eng, int) and d_eps_app_eng == 0:
            d_eps_app_eng = zeros_like(eps_app_eng)

        D_mtx = zeros(
            (eps_app_eng.shape[0], eps_app_eng.shape[0]), dtype='float_')

        mats_state_array = sctx.mats_state_array

        for i, mats in enumerate(self._mats_list):

            # extract the stress components
            eps, d_eps = eps_app_eng[i], d_eps_app_eng[i]
            size = self._state_sizes[i]
            offset = self._state_offsets[i]
            sctx.mats_state_array = mats_state_array[offset: offset + size]

            sig_app_eng[i], D_mtx[i, i] = mats.get_corr_pred(
                sctx, eps, d_eps, tn, tn1)

        sctx.mats_state_array = mats_state_array

        return sig_app_eng, D_mtx

    def get_sig1(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[0:1]

    def get_sig2(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[3:]

    def get_shear_flow(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[1:2]

    def get_cohesive_stress(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[2:3]

    rte_dict = Property

    def _get_rte_dict(self):

        rte_dict = {}
        ix_maps = [0, 1, 2, 3]
        for name, mats, ix_map, size, offset in \
            zip(self._mats_names, self._mats_list, ix_maps, self._state_sizes,
                self._state_offsets):
            for key, v_eval in list(mats.rte_dict.items()):

                __call_v_eval = RTE1D5Bond(v_eval=v_eval,
                                           name=name + '_' + key,
                                           size=size,
                                           offset=offset,
                                           ix_map=ix_map)

                rte_dict[name + '_' + key] = __call_v_eval

        # sigma is also achievable through phase1_sig_app and phase_2_sig_app
        extra_rte_dict = {'sig1': self.get_sig1,
                          'sig2': self.get_sig2,
                          'shear_flow': self.get_shear_flow,
                          'cohesive_stress': self.get_cohesive_stress,
                          }
        rte_dict.update(extra_rte_dict)
        return rte_dict
    #-------------------------------------------------------------------------
    # Methods required by the mats_explore tool
    #-------------------------------------------------------------------------

    def new_cntl_var(self):
        return zeros(4, 'float_')

    def new_resp_var(self):
        return zeros(4, 'float_')


class RTE1D5Bond(HasTraits):

    v_eval = Callable
    name = String
    size = Int
    offset = Int
    ix_map = Int

    def __call__(self, sctx, u, *args, **kw):
        u_x = array([u[self.ix_map]], dtype='float')
        # save the spatial context
        mats_state_array = sctx.mats_state_array
        sctx.mats_state_array = mats_state_array[
            self.offset: self.offset + self.size]
        result = self.v_eval(sctx, u_x, *args, **kw)
        # put the spatial context back
        sctx.mats_state_array = mats_state_array

        return result


if __name__ == '__main__':

    mats = MATS1D5Bond()
    mats.configure_traits()



'''
Created on Feb 11, 2018

@author: rch
'''

import os
from ibvpy.mathkit.tensor import EPS

from mayavi import mlab
from tvtk.api import tvtk
from ibvpy.view.plot3d.viz3d import Vis3D, Viz3D

import numpy as np
import traits.api as tr


class Vis3DLattice(Vis3D):
    r'''Record the response triggered by the hist object the time increments
    upon a record event.
    '''
    var = tr.Str('<unnamed>')

    def setup(self):
        self.new_dir()
        fe_domain = self.sim.tstep.fe_domain
        X_Ia_list = []
        I_Li_list = []
        Xm_Lia_list = []
        # Consistently with other adaptors, this one allows for
        # a loop over several lattice domains.
        for domain in fe_domain:
            xdomain = domain.xmodel
            var_function = domain.tmodel.var_dict.get(self.var, None)
            if var_function == None or xdomain.hidden:
                continue
            DELTA_x_ab = xdomain.vtk_expand_operator
            X_Ia = np.einsum(
                'Ia,ab->Ib',
                xdomain.X_Ia, DELTA_x_ab
            )
            Xm_Lia = np.einsum(
                'Lib,ba->Lia',
                xdomain.Xm_Lia, DELTA_x_ab
            )
            X_Ia_list.append(X_Ia)
            Xm_Lia_list.append(Xm_Lia)
            I_Li_list.append(xdomain.I_Li)
        if len(I_Li_list) == 0:
            raise ValueError(
                'Empty output for field variable %s in model %s' %
                (self.var, domain.tmodel)
            )
        X_Ia = np.hstack(X_Ia_list)
        I_Li = np.hstack(I_Li_list)
        fname = 'lattice_tessellation'
        target_file = os.path.join(
            self.dir, fname.replace('.', '_')
        ) + '.npz'
        np.savez(target_file,
                 X_Ia=X_Ia, I_Li=I_Li, Xm_Lia=Xm_Lia)
        self.lattice_file = target_file

    def update(self):
        ts = self.sim.tstep
        U = ts.U_k
        t = ts.t_n1
        U_Ia_list = []
        U_Lipa_list = []
        state_fields = []
        for domain in ts.fe_domain:
            xdomain = domain.xmodel
            state_field = domain.state_n.get(self.var, None)
            if (xdomain.hidden):
                # If the state variable not present in the domain, skip
                continue
            if (state_field is not None):
                state_fields.append(state_field.flatten())
            DELTA_x_ab = xdomain.vtk_expand_operator
            U_Ib = U[xdomain.o_Ipa][..., 0, :]
            U_Ia_list.append(np.einsum(
                '...a,ab->...b', U_Ib, DELTA_x_ab
            ))
            U_Lipa = U[xdomain.o_Lipa]
            U_Lipa_list.append(np.einsum(
                '...a,ab->...b', U_Lipa, DELTA_x_ab
            ))
            if len(U_Lipa) == 0:
                raise ValueError('no fields for variable %s' % self.var)
        # Save the displacements
        U_Ia = np.vstack(U_Ia_list)
        U_Lipa = np.vstack(U_Lipa_list)
 #       state_vec = np.hstack(state_fields)
        fname = 'U_step_%008.4f' % t
        target_file = os.path.join(
            self.dir, fname.replace('.', '_')
        ) + '.npz'
        # , state_vec=state_vec)
        np.savez(target_file, U_Ia=U_Ia, U_Lipa=U_Lipa)
        self.add_file(target_file)


class Viz3DLattice(Viz3D):

    warp_factor = tr.Float(1.0, auto_set=False, enter_set=True)

    def setup(self):
        m = mlab

        npzfile = np.load(self.vis3d.lattice_file)
        X_Ia = npzfile['X_Ia']
        I_Li = npzfile['I_Li']
        Xm_Lia = npzfile['Xm_Lia']

        self.d = tvtk.PolyData(points=X_Ia, lines=I_Li)
        self.warp_vector = m.pipeline.warp_vector(self.d)
        self.warp_vector.filter.scale_factor = self.warp_factor

        tube = m.pipeline.tube(self.warp_vector,
                               tube_radius=self.tube_radius)
        self.lines = m.pipeline.surface(tube, color=(0.1, 0.1, 0.1))

        X_aiL = np.einsum('Lia->aiL', X_Ia[I_Li])
        Xm_aiL = np.einsum('Lia->aiL', Xm_Lia)
        XIm_aiL = np.concatenate(
            [
                np.einsum('iaL->aiL',
                          np.array([X_aiL[:, 0, ...], Xm_aiL[:, 0, ...]])),
                np.einsum('iaL->aiL',
                          np.array([X_aiL[:, 1, ...], Xm_aiL[:, 0, ...]]))
            ], axis=-1
        )
        XIm_aLi = np.einsum('aiL->aLi', XIm_aiL)
        XIm_aJ = XIm_aLi.reshape(3, -1)
        XIm_Ja = np.einsum('aJ->Ja', XIm_aJ)
        J_Li = np.arange(XIm_aJ.shape[1]).reshape(-1, 2)
        self.d_J = tvtk.PolyData(points=XIm_Ja, lines=J_Li)
        self.warp_vector_J = m.pipeline.warp_vector(self.d_J)
        self.warp_vector_J.filter.scale_factor = self.warp_factor

        tube_J = m.pipeline.tube(self.warp_vector_J,
                                 tube_radius=self.tube_radius)
        self.lines_J = m.pipeline.surface(tube_J, color=(0.1, 0.1, 0.1))
        glyph = m.pipeline.glyph(self.warp_vector_J)

        glyph.glyph.glyph_source.glyph_source = glyph.glyph.glyph_source.glyph_dict[
            'cube_source']
        glyph.glyph.glyph.range = np.array([0., 0.1])
        glyph.glyph.scale_mode = 'scale_by_vector'
        glyph.glyph.color_mode = 'color_by_vector'

    tube_radius = tr.Float(0.01)

    visible = tr.Property(tr.Bool)

    def _set_visible(self, visible):
        return
        self.d.visible = visible

    def _get_visible(self):
        return True  # self.d.visible

    def plot(self, vot):
        fe_domain = self.vis3d.sim.tstep.fe_domain
        xdomain = fe_domain[0].xmodel
        dXm_Lia = xdomain.dXm_Lia

        idx = self.vis3d.sim.hist.get_time_idx(vot)
        fname = self.vis3d.file_list[idx]
        npzfile = np.load(fname)
        U_Ia = npzfile['U_Ia']
        self.d.point_data.vectors = U_Ia

        U_Lipa = npzfile['U_Lipa']
        Um_trans_Lia, Phi_Lia = np.einsum('Lipa->pLia', U_Lipa)
        Um_rot_Lia = np.einsum('abc,...b,...c->...a',
                               EPS, Phi_Lia, dXm_Lia)
        Um_Lia = Um_trans_Lia + Um_rot_Lia
        Um_trans_aiL = np.einsum('Lia->aiL', Um_trans_Lia)
        Um_aiL = np.einsum('Lia->aiL', Um_Lia)
        UIm_aiL = np.concatenate(
            [
                np.einsum('iaL->aiL',
                          np.array([Um_trans_aiL[:, 0, ...], Um_aiL[:, 0, ...]])),
                np.einsum('iaL->aiL',
                          np.array([Um_trans_aiL[:, 1, ...], Um_aiL[:, 1, ...]]))
            ], axis=-1
        )
        UIm_aLi = np.einsum('aiL->aLi', UIm_aiL)
        UIm_aJ = UIm_aLi.reshape(3, -1)
        UIm_Ja = np.einsum('aJ->Ja', UIm_aJ)
        self.d_J.point_data.vectors = UIm_Ja



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Mar 2, 2010 by: rch


from matplotlib.figure import Figure
from numpy import linspace, frompyfunc
from traits.api import Float, \
    Instance, on_trait_change, Int, \
    Event
from traitsui.api import \
    View, Item, VGroup, HGroup, ModelView, HSplit, VSplit
from traitsui.menu import \
    OKButton
from util.traits.editors.mpl_figure_editor import MPLFigureEditor

from .mats1D_damage import MATS1DDamage


class MATS1DDamageView(ModelView):
    '''
    View into the parametric space constructed over the model.

    The is associated with the PStudySpace instance covering the
    factor ranges using an n-dimensional array.

    The view is responsible for transferring the response values
    into 2D and 3D plots. Depending on the current view specification
    it also initiates the calculation of response values in the 
    currently viewed subspace of the study. 
    '''

    model = Instance(MATS1DDamage)

    max_eps = Float(0.01,
                    enter_set=True,
                    auto_set=False,
                    modified=True)

    n_points = Int(100,
                   enter_set=True,
                   auto_set=False,
                   modified=True)

    data_changed = Event(True)

    @on_trait_change('+modified,model.+modified')
    def _redraw(self):

        get_omega = frompyfunc(self.model._get_omega, 1, 1)
        xdata = linspace(0, self.max_eps, self.n_points)
        ydata = get_omega(xdata)

        axes = self.figure.axes[0]
        axes.clear()

        axes.plot(xdata, ydata, color='blue', linewidth=3)

        self.data_changed = True

    #---------------------------------------------------------------
    # PLOT OBJECT
    #-------------------------------------------------------------------
    figure = Instance(Figure)

    def _figure_default(self):
        figure = Figure(facecolor='white')
        figure.add_axes([0.12, 0.13, 0.85, 0.74])
        return figure

    traits_view = View(HSplit(
        VGroup(
            Item('model@', show_label=False, resizable=True),
            label='material parameters',
            id='tmodel.viewmodel.model',
            dock='tab',
        ),
        VSplit(
            VGroup(
                Item('figure',  editor=MPLFigureEditor(),
                     resizable=True, show_label=False),
                label='plot sheet',
                id='tmodel.viewmodel.figure_window',
                dock='tab',
            ),
            VGroup(
                HGroup(
                    Item('max_eps', label='maximum epsilon [-]',
                         springy=True),
                    Item('n_points', label='plot points',
                         springy=True),
                ),
                label='plot parameters',
                id='tmodel.viewmodel.view_params',
                dock='tab',
            ),
            id='tmodel.viewmodel.right',
        ),
        id='tmodel.viewmodel.splitter',
    ),
        title='Yarn Size Effect',
        id='yse.viewmodel',
        dock='tab',
        resizable=True,
        height=0.8, width=0.8,
        buttons=[OKButton])

if __name__ == '__main__':

    mats_eval = MATS1DDamageView(model=MATS1DDamage())
    mats_eval.configure_traits()



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 3, 2009 by jakub

from ibvpy.tmodel.mats_eval import MATSEval
from traits.api import Callable, Constant


def identity_mapping(var):
    return var


class MATS1D5Eval(MATSEval):

    # number of physical dimensions with defined integration
    #
    n_dims = Constant(1)


class NotUsed:
    # dimension-dependent mappings
    #
    map_tns4_to_tns2 = Callable(identity_mapping, transient=True)
    map_eps_eng_to_mtx = Callable(identity_mapping, transient=True)
    map_sig_eng_to_mtx = Callable(identity_mapping, transient=True)
    compliance_mapping = Callable(identity_mapping, transient=True)
    map_sig_mtx_to_eng = Callable(identity_mapping, transient=True)
    map_eps_mtx_to_eng = Callable(identity_mapping, transient=True)




from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from ibvpy.tmodel.mats_damage_fn import \
    GfDamageFn
from traits.api import on_trait_change, Bool
from traitsui.api import UItem, \
    Item, View, Group, Spring

import numpy as np
import traits.api as tr

from .vstrain_norm2d import StrainNorm2D, SN2DRankine


class MATS2DScalarDamage(MATS2DEval):
    r'''Isotropic damage model.
    '''

    node_name = 'isotropic damage model'

    tree_node_list = tr.List

    def _tree_node_list_default(self):
        return [self.strain_norm, self.omega_fn, ]

    def _update_node_list(self):
        self.tree_node_list = [
            self.strain_norm,
            self.omega_fn,
        ]

    stiffness = tr.Enum("secant", "algorithmic",
                        input=True)
    r'''Selector of the stiffness calculation.
    '''
    #=========================================================================
    # Material model
    #=========================================================================
    strain_norm_type = tr.Trait('Rankine',
                                {'Rankine': SN2DRankine,
                                 },
                                MAT=True
                                )

    @on_trait_change('strain_norm_type')
    def _set_strain_norm(self):
        self.strain_norm = self.strain_norm_type_(mats=self)
        self._update_node_list()

    strain_norm = tr.Instance(StrainNorm2D,
                              report=True)
    '''Material model'''

    def _strain_norm_default(self):
        return self.strain_norm_type_(mats=self)

    omega_fn = tr.Instance(GfDamageFn)

    def _omega_fn_default(self):
        return GfDamageFn(mats=self)

    changed = tr.Event
    r'''This event can be used by the clients to trigger 
    an action upon the completed reconfiguration of 
    the material model
    '''

    state_var_shapes = {'kappa': (),
                        'omega': ()}
    r'''
    Shapes of the state variables
    to be stored in the global array at the level 
    of the domain.
    '''

    def init(self, kappa, omega):
        r'''
        Initialize the state variables.
        '''
        kappa[...] = 0
        omega[...] = 0

    algorithmic = Bool(True)

    def get_corr_pred(self, eps_ab, tn1, kappa, omega):
        r'''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        eps_eq = self.strain_norm.get_eps_eq(eps_ab, kappa)
        I = self.omega_fn.get_f_trial(eps_eq)
        eps_eq_I = eps_eq[I]
        kappa[I] = eps_eq_I
        omega[I] = self.omega_fn(eps_eq_I)
        phi = (1.0 - omega)
        D_abcd = np.einsum(
            '...,abcd->...abcd',
            phi, self.D_abcd
        )
        sig_ab = np.einsum(
            '...abcd,...cd->...ab',
            D_abcd, eps_ab
        )
        if self.algorithmic:
            domega_I = self.omega_fn.diff(eps_eq_I)
            domega_I[np.where(np.fabs(domega_I) < .001)] = .001
            deps_eq_cd_I = self.strain_norm.get_deps_eq(eps_ab[I])
            D_abcd[I] -= np.einsum(
                '...,...cd,abcd,...cd->...abcd',
                domega_I, deps_eq_cd_I, self.D_abcd, eps_ab[I]
            )
        return sig_ab, D_abcd

    def _get_var_dict(self):
        var_dict = super(MATS2DScalarDamage, self)._get_var_dict()
        var_dict.update(
            omega=self.get_omega
        )
        return var_dict

    def get_omega(self, eps_ab, tn1, kappa, omega):
        return omega

    traits_view = View(
        Group(
            Group(
                Item('E', resizable=True),
                Item('nu', resizable=True),
                Item('strain_norm_type', resizable=True),
                Group(
                    UItem('omega_fn@', full_size=True),
                    label='Damage function'
                )
            ),
            Group(
                Item('stress_state', style='custom'),
                Item('stiffness', style='custom'),
                Spring(resizable=True),
                label='Configuration parameters',
                show_border=True,
            ),
        ),
        resizable=True
    )

    tree_view = traits_view


if __name__ == '__main__':

    #-------------------------------------------------------------------------
    # Example
    #-------------------------------------------------------------------------

    f_t = 2.4
    G_f = 0.090
    E = 20000.0
    omega_fn_gf = GfDamageFn(G_f=G_f, f_t=f_t, L_s=22.5)
    print(omega_fn_gf.eps_0)
    print(omega_fn_gf(0.000148148))

    mats = MATS2DScalarDamage(E=30000,
                              stiffness='algorithmic',
                              nu=0.0,
                              )
    print(mats.strain_norm)
    mats.configure_traits()




from os.path import join

from ibvpy.mathkit.mfn.mfn_line.mfn_line import MFnLineArray
from ibvpy.view.reporter import RInputRecord
from scipy.optimize import newton
from traits.api import \
    Instance, Str, \
    Float, on_trait_change,\
    Interface, provides, Range, Property, Button, \
    Array, WeakRef
from traitsui.api import \
    View, Item, UItem, VGroup
from ibvpy.view.ui import BMCSLeafNode

import numpy as np


class PlottableFn(RInputRecord):

    plot_min = Float(0.0, input=True,
                     enter_set=True, auto_set=False)
    plot_max = Float(1.0, input=True,
                     enter_set=True, auto_set=False)

    fn = Instance(MFnLineArray)

    def _fn_default(self):
        return MFnLineArray()

    def __init__(self, *args, **kw):
        super(PlottableFn, self).__init__(*args, **kw)
        self.update()

    def plot(self, ax, **kw):
        n_vals = 200
        xdata = np.linspace(self.plot_min, self.plot_max, n_vals)
        ydata = np.zeros_like(xdata)
        f_idx = self.get_f_trial(xdata)
        ydata[f_idx] = self.__call__(xdata[f_idx])
        color = kw.pop('color', 'green')
        ax.plot(xdata, ydata, color=color, **kw)

    @on_trait_change('+input')
    def update(self):
        n_vals = 200
        xdata = np.linspace(self.plot_min, self.plot_max, n_vals)
        ydata = np.zeros_like(xdata)
        f_idx = self.get_f_trial(xdata)
        if len(f_idx) > 0:
            ydata[f_idx] = self.__call__(xdata[f_idx])
        self.fn.set(xdata=xdata, ydata=ydata)
        self.fn.replot()

    def write_figure(self, f, rdir, rel_study_path):
        fname = 'fig_' + self.node_name.replace(' ', '_') + '.pdf'
        f.write(r'''
\multicolumn{3}{r}{\includegraphics[width=5cm]{%s}}\\
''' % join(rel_study_path, fname))
        self.fn.savefig(join(rdir, fname))

    traits_view = View(UItem('fn'))


class IDamageFn(Interface):

    def __call__(self, k):
        '''get the value of the function'''

    def diff(self, k):
        '''get the first derivative of the function'''

    def get_f_trial(self, k):
        '''get the map of indexes with inelastic behavior'''


class DamageFn(BMCSLeafNode, PlottableFn):

    mats = WeakRef

    s_0 = Float(0.0004,
                MAT=True,
                input=True,
                symbol="s_0",
                desc="elastic strain limit",
                unit='mm',
                enter_set=True,
                auto_set=False)

    def diff(self, k):
        return self.fn.diff(k)

    latex_eq = Str(None)

    def _repr_latex_(self):
        return self.latex_eq + super(DamageFn, self)._repr_latex_()


class GfDamageFn(DamageFn):
    '''Class defining the damage function coupled with the fracture 
    energy of a cohesive crack model.
    '''
    node_name = 'damage function Gf'

    L_s = Float(1.0,
                MAT=True,
                input=True,
                label="L_s",
                desc="Length of the softening zone",
                enter_set=True,
                auto_set=False)

    E_ = Float(34000.0,
               MAT=True,
               input=True,
               label="E",
               desc="Young's modulus",
               enter_set=True,
               auto_set=False)

    E = Property()

    def _get_E(self):
        if self.mats:
            return self.mats.E
        else:
            return self.E_

    f_t = Float(4.5,
                MAT=True,
                input=True,
                label="f_t",
                desc="Tensile strength",
                enter_set=True,
                auto_set=False)

    f_t_Em = Array(np.float_, value=None)

    G_f = Float(0.004,
                MAT=True,
                input=True,
                label="G_f",
                desc="Fracture energy",
                enter_set=True,
                auto_set=False)

    eps_0 = Property()

    def _get_eps_0(self):
        return self.f_t / self.E

    eps_ch = Property()

    def _get_eps_ch(self):
        return self.G_f / self.f_t

    plot_max = Property(depends_on='G_f,f_t,E')

    def _get_plot_max(self):
        return self.eps_ch * self.L_s * 3.0

    def get_f_trial(self, eps_eq_Em):
        f_t = self.f_t
        if len(self.f_t_Em) > 0:
            f_t = self.f_t * self.f_t_Em
        eps_0 = f_t / self.E
        return np.where(eps_eq_Em - eps_0 > 0)

    def __call__(self, kappa):
        L_s = self.L_s
        f_t = self.f_t
        G_f = self.G_f
        E = self.E
        eps_0 = self.eps_0
        return (
            1 - f_t * np.exp(-f_t * (kappa - eps_0) * L_s / G_f)
            / (E * kappa)
        )

    def diff(self, kappa):
        L_s = self.L_s
        f_t = self.f_t
        G_f = self.G_f
        E = self.E
        eps_0 = self.eps_0
        return (
            f_t * np.exp(L_s * (eps_0 - kappa) * f_t / G_f)
            / (E * G_f * kappa**2) * (G_f + L_s * kappa * f_t)
        )

    traits_view = View(
        VGroup(
            VGroup(
                Item('f_t'),
                Item('G_f'),
                Item('E', style='readonly'),
                Item('L_s', style='readonly'),
                Item('s_0', style='readonly',
                     full_size=True, resizable=True),
            ),
            VGroup(
                UItem('fn@', height=300)
            )
        )
    )

    tree_view = traits_view


@provides(IDamageFn)
class JirasekDamageFn(DamageFn):

    node_name = 'Jirasek damage function'

    s_f = Float(0.001,
                MAT=True,
                input=True,
                symbol="s_\mathrm{f}",
                unit='mm/mm',
                desc="derivative of the damage function at the onset of damage",
                enter_set=True,
                auto_set=False)

    plot_max = 1e-2

    def get_f_trial(self, eps_eq_Em):
        eps_0 = self.s_0
        return np.where(eps_eq_Em - eps_0 > 0)

    def __call__(self, kappa):
        s_0 = self.s_0
        s_f = self.s_f
        omega = np.zeros_like(kappa, dtype=np.float_)
        I = np.where(kappa >= s_0)
        k_I = kappa[I]
        omega[I] = 1. - s_0 / k_I * np.exp(-1 * (k_I - s_0) / (s_f - s_0))
        return omega

    def diff(self, kappa):
        s_0 = self.s_0
        s_f = self.s_f
        I = np.where(kappa >= s_0)
        k_I = kappa[I]
        domega_dkappa = np.zeros_like(kappa)
        domega_dkappa[I] = (
            s_0 * (-k_I + s_0 - s_f) * np.exp((k_I - s_0) /
                                              (s_0 - s_f)) / (k_I**2 * (s_0 - s_f))
        )
        return domega_dkappa

    latex_eq = Str(r'''Damage function (Jirasek)
        \begin{align}
        \omega &= g(\kappa) 
        = 1 - \left[\frac{s_0}{\kappa} \exp \left(- \frac{\kappa 
        - s_0}{s_f - s_0}\right)\right]
        \end{align}
        where $\kappa$ is the state variable representing 
        the maximum slip that occurred so far in
        in the history of loading.
        ''')

    traits_view = View(
        VGroup(
            VGroup(
                Item('s_0', full_size=True, resizable=True),
                Item('s_f'),
                Item('plot_max'),
            ),
            VGroup(
                UItem('fn@', height=300)
            )
        )
    )

    tree_view = traits_view


@provides(IDamageFn)
class LiDamageFn(DamageFn):

    node_name = 'Li damage function'

    latex_eq = Str(r'''Damage function (Li)
        \begin{align}
        \omega = g(\kappa) = \frac{\alpha_1}{1 + \exp(-\alpha_2 \kappa + 6 )}
        \end{align}
        where $\kappa$ is the state variable representing 
        the maximum slip that occurred so far in
        in the history of loading.
        ''')

    alpha_1 = Range(value=1., low=0.0, high=1.0,
                    MAT=True,
                    input=True,
                    symbol=r'\alpha_1',
                    unit='-',
                    desc="parameter controlling the shape of the damage function",
                    enter_set=True,
                    auto_set=False)

    alpha_2 = Float(2000.,
                    MAT=True,
                    input=True,
                    symbol=r'\alpha_2',
                    unit='-',
                    desc="parameter controlling the shape of the damage function",
                    enter_set=True,
                    auto_set=False)

    plot_max = 1e-2

    def get_f_trial(self, eps_eq_Em):
        eps_0 = self.s_0
        return np.where(eps_eq_Em - eps_0 > 0)

    def __call__(self, kappa):
        alpha_1 = self.alpha_1
        alpha_2 = self.alpha_2
        s_0 = self.s_0
        omega = np.zeros_like(kappa, dtype=np.float_)
        d_idx = np.where(kappa >= s_0)[0]
        k = kappa[d_idx]
        omega[d_idx] = 1. / \
            (1. + np.exp(-1. * alpha_2 * (k - s_0) + 6.)) * alpha_1
        return omega

    def diff(self, kappa):
        alpha_1 = self.alpha_1
        alpha_2 = self.alpha_2
        s_0 = self.s_0
        return ((alpha_1 * alpha_2 *
                 np.exp(-1. * alpha_2 * (kappa - s_0) + 6.)) /
                (1 + np.exp(-1. * alpha_2 * (kappa - s_0) + 6.)) ** 2)

    traits_view = View(
        VGroup(
            VGroup(
                Item('s_0', style='readonly', full_size=True, resizable=True),
                Item('alpha_1'),
                Item('alpha_2'),
                Item('plot_max'),
            ),
            VGroup(
                UItem('fn@', height=300)
            )
        )
    )

    tree_view = traits_view


@provides(IDamageFn)
class AbaqusDamageFn(DamageFn):

    node_name = 'Abaqus damage function'

    latex_eq = Str(r'''Damage function (Abaqus)
        \begin{align}
        \omega = g(\kappa) = 1 -\left(\frac{s_0}{\kappa}\right)\left[ 1 - \frac{1 - \exp(- \alpha(\frac{\kappa - s_0}{s_u - s_0})}{1 - \exp(-\alpha)}  \right]
        \end{align}
        where $\kappa$ is the state variable representing 
        the maximum slip that occurred so far in
        in the history of loading.
        ''')

    s_u = Float(0.003,
                MAT=True,
                input=True,
                symbol="s_u",
                unit='mm',
                desc="parameter of the damage function",
                enter_set=True,
                auto_set=False)

    alpha = Float(0.1,
                  MAT=True,
                  input=True,
                  symbol=r"\alpha",
                  desc="parameter controlling the slope of damage",
                  unit='-',
                  enter_set=True,
                  auto_set=False)

    plot_max = 1e-3

    def get_f_trial(self, kappa):
        s_0 = self.s_0
        return np.where(kappa > s_0)[0]

    def __call__(self, kappa):
        s_0 = self.s_0
        s_u = self.s_u
        alpha = self.alpha

        omega = np.zeros_like(kappa, dtype=np.float_)
        d_idx = np.where(kappa > s_0)[0]
        k = kappa[d_idx]

        sk = (k - s_0) / (s_u - s_0)
        frac = (1 - np.exp(-alpha * sk)) / (1 - np.exp(-alpha))

        omega[d_idx] = 1 - s_0 / k * (1 - frac)
        omega[np.where(omega > 1.0)] = 1.0
        return omega

    def diff(self, kappa):
        s_0 = self.s_0
        s_u = self.s_u
        alpha = self.alpha
        denom = ((np.exp(alpha) - 1) * kappa**2 * (s_u - s_0))
        denom[np.where(np.fabs(denom) < 1e-5)] = 1e-8
        d_g_eps = (
            - s_0 * np.exp(alpha * (kappa - s_0) / (s_u - s_0))
            * (
                (s_u - s_0) * np.exp(alpha * (kappa - s_0) / (s_u - s_0))
                + np.exp(alpha) * (alpha * kappa - s_u + s_0)
            )
            / denom
        )
        d_g_eps[np.where(kappa - s_0 < 0)] = 0.0
        return d_g_eps

    traits_view = View(
        VGroup(
            VGroup(
                Item('s_0',
                     full_size=True, resizable=True),
                Item('s_u'),
                Item('alpha'),
                Item('plot_max'),
            ),
            VGroup(
                UItem('fn@', height=300)
            )
        )
    )

    tree_view = traits_view


@provides(IDamageFn)
class FRPDamageFn(DamageFn):

    node_name = 'FRP damage function'

    B = Float(10.4,
              MAT=True,
              input=True,
              symbol="B",
              unit='mm$^{-1}$',
              desc="parameter controlling the damage maximum stress level",
              enter_set=True,
              auto_set=False)

    Gf = Float(1.19,
               MAT=True,
               input=True,
               symbol="G_\mathrm{f}",
               unit='N/mm',
               desc="fracture energy",
               enter_set=True,
               auto_set=False)

    plot_max = 0.5

    def __init__(self, *args, **kw):
        super(FRPDamageFn, self).__init__(*args, **kw)
        self._update_dependent_params()

    E_bond = Float(0.0)

    E_b = Property(Float)

    def _get_E_b(self):
        return self.mats.E_b

    def _set_E_b(self, value):
        self.E_bond = value
        self.mats.E_b = value

    @on_trait_change('B, Gf')
    def _update_dependent_params(self):
        self.E_b = 1.734 * self.Gf * self.B ** 2.0
        # calculation of s_0, implicit function solved using Newton method

        def f_s(s_0): return s_0 / \
            (np.exp(- self.B * s_0) - np.exp(-2.0 * self.B * s_0)) - \
            2.0 * self.B * self.Gf / self.E_b
        self.s_0 = newton(f_s, 0.00000001, tol=1e-5, maxiter=20)

    def get_f_trial(self, eps_eq_Em):
        eps_0 = self.s_0
        return np.where(eps_eq_Em - eps_0 > 0)

    def __call__(self, kappa):

        b = self.B
        Gf = self.Gf
        Eb = self.E_b  # 1.734 * Gf * b**2
        s_0 = self.s_0
        # calculation of s_0, implicit function solved using Newton method

#         def f_s(s_0): return s_0 / \
#             (np.exp(-b * s_0) - np.exp(-2.0 * b * s_0)) - 2.0 * b * Gf / Eb
#         s_0 = newton(f_s, 0.00000001, tol=1e-5, maxiter=20)

        omega = np.zeros_like(kappa, dtype=np.float_)
        I = np.where(kappa >= s_0)[0]
        kappa_I = kappa[I]

        omega[I] = 1 - \
            (2.0 * b * Gf * (np.exp(-b * kappa_I)
                             - np.exp(-2.0 * b * kappa_I))) / (kappa_I * Eb)

        return omega

    def diff(self, kappa):

        nz_ix = np.where(kappa != 0.0)[0]

        b = self.B
        Gf = self.Gf
        Eb = 1.734 * Gf * b**2

        domega_dkappa = np.zeros_like(kappa)
        kappa_nz = kappa[nz_ix]
        domega_dkappa[nz_ix] = (
            (2.0 * b * Gf *
             (np.exp(-b * kappa_nz) -
              np.exp(-2.0 * b * kappa_nz))
             ) / (Eb * kappa_nz**2.0) -
            (2.0 * b * Gf *
             (-b * np.exp(-b * kappa_nz) +
              2.0 * b * np.exp(-2.0 * b * kappa_nz))) /
            (Eb * kappa_nz)
        )
        return domega_dkappa

    latex_eq = r'''Damage function (FRP)
        \begin{align}
        \omega = g(\kappa) = 
        1 - {\frac {{\exp(-2\,Bs)}-{\exp(-Bs)}}{Bs}}
        \end{align}
        where $\kappa$ is the state variable representing 
        the maximum slip that occurred so far in
        in the history of loading.
        '''

    traits_view = View(
        VGroup(
            VGroup(
                Item('s_0', style='readonly', full_size=True, resizable=True),
                Item('E_bond', style='readonly',
                     full_size=True, resizable=True),
                Item('B'),
                Item('Gf'),
                Item('plot_max'),
            ),
            VGroup(
                UItem('fn@', height=300)
            )
        )
    )

    tree_view = traits_view


@provides(IDamageFn)
class MultilinearDamageFn(DamageFn):

    node_name = 'Multilinear damage function'

    s_data = Str('0,1', tooltip='Comma-separated list of strain values',
                 MAT=True, unit='mm', symbol='s',
                 desc='slip values',
                 auto_set=True, enter_set=False)

    omega_data = Str('0,0', tooltip='Comma-separated list of damage values',
                     MAT=True, unit='-', symbol=r'\omega',
                     desc='shear stress values',
                     auto_set=True, enter_set=False)

    def get_f_trial(self, eps_eq_Em):
        eps_0 = self.damage_law.xdata[1]
        return np.where(eps_eq_Em - eps_0 > 0)[0]

    s_omega_table = Property

    def _set_s_omega_table(self, data):
        s_data, omega_data = data
        if len(s_data) != len(omega_data):
            raise ValueError('s array and tau array must have the same size')
        self.damage_law.set(xdata=s_data,
                            ydata=omega_data)

    update_damage_law = Button(label='update bond-slip law')

    def _update_damage_law_fired(self):
        s_data = np.fromstring(self.s_data, dtype=np.float_, sep=',')
        omega_data = np.fromstring(self.omega_data, dtype=np.float_, sep=',')
        if len(s_data) != len(omega_data):
            raise ValueError('s array and tau array must have the same size')
        self.damage_law.set(xdata=s_data,
                            ydata=omega_data)
        self.damage_law.replot()

    damage_law = Instance(MFnLineArray)

    def _damage_law_default(self):
        return MFnLineArray(
            xdata=[0.0, 1.0],
            ydata=[0.0, 0.0],
            plot_diff=False)

    def __call__(self, kappa):
        shape = kappa.shape
        return self.damage_law(kappa.flatten()).reshape(*shape)

    def diff(self, kappa):
        shape = kappa.shape
        return self.damage_law.diff(kappa.flatten()).reshape(*shape)

    traits_view = View(
        VGroup(
            VGroup(
                Item('s_data', full_size=True, resizable=True),
                Item('omega_data'),
                UItem('update_damage_law')
            ),
            UItem('damage_law@')
        )
    )

    tree_view = traits_view


if __name__ == '__main__':
    #ld = LiDamageFn()
    #ld = GfDamageFn()
    #mld = MultilinearDamageFn()
    #mld = FRPDamageFn(Gf=100)
    mld = JirasekDamageFn()
    mld.configure_traits()




import copy

from ibvpy.tmodel.mats_eval import \
    MATSEval
from traits.api import \
    Constant, Float, Property, cached_property, Array,\
    Dict, Str, Callable

import numpy as np
import traitsui.api as ui


class MATSXDEval(MATSEval):
    '''Base class for elastic models.
    '''

    n_dims = Constant(Float)
    '''Number of spatial dimensions of an integration 
    cell for the material model
    '''

    E = Float(34e+3,
              label="E",
              desc="Young's Modulus",
              auto_set=False,
              input=True)

    nu = Float(0.2,
               label='nu',
               desc="Poison's ratio",
               auto_set=False,
               input=True)

    def _get_lame_params(self):
        # First Lame parameter (bulk modulus)
        la = self.E * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E / (2. + 2. * self.nu)
        return la, mu

    D_abef = Property(Array, depends_on='+input')
    '''Material stiffness - rank 4 tensor
    '''
    @cached_property
    def _get_D_abef(self):
        la, mu = self._get_lame_params()
        delta = np.identity(self.n_dims)
        return (
            np.einsum(',ij,kl->ijkl', la, delta, delta) +
            np.einsum(',ik,jl->ijkl', mu, delta, delta) +
            np.einsum(',il,jk->ijkl', mu, delta, delta)
        )

    def get_corr_pred(self, eps_Emab, tn1, **kw):
        '''
        Corrector predictor computation.
        @param eps_Emab input variable - strain tensor
        '''
        sigma_Emab = np.einsum(
            'abcd,...cd->...ab', self.D_abef, eps_Emab
        )
        Em_len = len(eps_Emab.shape) - 2
        new_shape = tuple([1 for _ in range(Em_len)]) + self.D_abef.shape
        D_abef = self.D_abef.reshape(*new_shape)
        return sigma_Emab, D_abef

    #=========================================================================
    # Response variables
    #=========================================================================
    def get_eps_ab(self, eps_ab, tn1, **Eps):
        return eps_ab

    def get_sig_ab(self, eps_ab, tn1, **Eps):
        return self.get_sig(eps_ab, tn1, **Eps)

    var_dict = Property(Dict(Str, Callable))
    '''Dictionary of response variables
    '''
    @cached_property
    def _get_var_dict(self):
        return dict(eps_ab=self.get_eps_ab,
                    sig_ab=self.get_sig_ab)

    def x_var_dict_default(self):
        return {'sig_app': self.get_sig_ab,
                'eps_app': self.get_eps_ab,
                'max_principle_sig': self.get_max_principle_sig,
                'strain_energy': self.get_strain_energy}

    view_traits = ui.View(
        ui.VSplit(
            ui.Group(
                ui.Item('E'),
                ui.Item('nu'),),
        ),
        resizable=True
    )




from ibvpy.api import BCDof
from ibvpy.api import TLoop, TLine
from ibvpy.tmodel.mats2D.mats2D_cmdm.mats2D_cmdm import MATS2DMicroplaneDamage
from ibvpy.tmodel.mats2D.mats2D_elastic.vmats2D_elastic import MATS2DElastic
from ibvpy.tmodel.mats2D.mats2D_sdamage.vmats2D_sdamage import MATS2DScalarDamage
from ibvpy.tmodel.matsXD.matsXD_explore import MATSXDExplore
from ibvpy.tmodel.mats_eval import IMATSEval
from ibvpy.mesh.fe_domain import FEDomain
from ibvpy.mesh.fe_grid import FEGrid
from traits.api import \
     Array, Bool, Callable, Enum, Float, HasTraits, \
     Instance, Int, Trait, Range, HasStrictTraits, on_trait_change, Event, \
     Dict, Property, cached_property, Delegate, List, WeakRef
from traitsui.api import \
    Item, View, HSplit, VSplit, VGroup, Group, Spring
from ibvpy.util.traits.either_type import \
   EitherType


class MATS2DExplore(MATSXDExplore):
    '''
    Simulate the loading histories of a material point in 2D space.
    '''

    mats_eval = EitherType(klasses=[MATS2DElastic,
                                    MATS2DMicroplaneDamage,
                                    MATS2DScalarDamage ])

    def _mats_eval_default(self):
        return MATS2DElastic()




from .sim.sim_base import Simulator
from .sim.hist import Hist
from .sim.i_hist import IHist
from .sim.i_tmodel import ITModel
from .sim.i_simulator import ISimulator
from .sim.i_tloop import ITLoop
from .sim.i_tstep import ITStep
from .sim.i_xmodel import IXModel
from .sim.tmodel import TModel
from .sim.tline import TLine
from .sim.tloop import TLoop
from .sim.tstep import TStep
from .sim.tstep_bc import TStepBC
from .xmodel.xdomain_fe_grid import XDomainFEGrid
from .xmodel.xdomain_interface1d import XDomainFEInterface1D
from .xmodel.xdomain_lattice import XDomainLattice
from .xmodel.xdomain_point import XDomainSinglePoint
from .bcond.bc_dof import BCDof
from .bcond.bc_dofgroup import BCDofGroup
from .bcond.bc_slice import BCSlice
from .fets.fets import FETSEval
from .fets.i_fets import IFETSEval
from .tmodel import MATSEval, IMATSEval
from .tmodel import MATSBondSlipMultiLinear, MATS1D5D
from .mesh.fe_domain import FEDomain
from .mesh.fe_grid import FEGrid
from .mesh.fe_grid_idx_slice import FEGridIdxSlice
from .mesh.fe_grid_ls_slice import FEGridLevelSetSlice
from .mesh.fe_refinement_grid import FERefinementGrid, FERefinementGrid as FEPatchedGrid



#-- Imports --------------------------------------------------------------

from numpy \
    import sqrt
from numpy.random \
    import random
from traits.api \
    import HasTraits, Property, Array, Any, Event, \
    on_trait_change, Instance, WeakRef, Int, Str, Bool, Trait, \
    Interface, provides
from traitsui.api \
    import View, Item, TabularEditor, HSplit, Group
from traitsui.menu \
    import CancelButton
from traitsui.tabular_adapter \
    import TabularAdapter


#------------------------------------------------------------------------
# Source of the data for array view
#------------------------------------------------------------------------
class ICellArraySource(Interface):

    '''Object representing a structured 1,2,3-dimensional cell grid.
    '''
    pass

#------------------------------------------------------------------------
# Cell view interface
#------------------------------------------------------------------------


class ICellView(Interface):

    '''Interface of the general cell view.
    '''

    def set_cell_traits(self):
        '''Adapt the view to the newly set cell_idx.
        '''
        raise NotImplementedError

    def redraw(self):
        '''Redraw the graphical representation of the cell 
        in the mayavi pipeline.
        '''
        raise NotImplementedError

#------------------------------------------------------------------------
# Default implementation of the cell view
#------------------------------------------------------------------------


@provides(ICellView)
class CellView(HasTraits):

    '''Get the element numbers.
    '''

    cell_idx = Int(-1)

    cell_grid = WeakRef(ICellArraySource)

    def set_cell(self, cell_idx):
        '''Method to be overloaded by subclasses. The subclass 
        can fetch the data required from the cell_grid 
        '''
        self.cell_idx = cell_idx
        self.set_cell_traits()

    def set_cell_traits(self):
        '''Specialize this function to fetch the cell data from 
        the array source.
        '''
        pass

    def redraw(self):
        '''No plotting defined by default'''
        pass

    view = View(Item('cell_idx'),
                resizable=True)

#-- Tabular Adapter Definition -------------------------------------------


class ElemTabularAdapter (TabularAdapter):

    columns = Property

    def _get_columns(self):
        data = getattr(self.object, self.name)
        if len(data.shape) != 2:
            raise ValueError('element node array must be two-dimensional')
        n_columns = getattr(self.object, self.name).shape[1]

        cols = [(str(i), i) for i in range(n_columns)]
        return [('element', 'index')] + cols

    font = 'Courier 10'
    alignment = 'right'
    format = '%d'
    index_text = Property
#    index_image = Property

    def _get_index_text(self):
        return str(self.row)

    def x_get_index_image(self):
        x, y, z = self.item
        if sqrt((x - 0.5) ** 2 + (y - 0.5) ** 2 + (z - 0.5) ** 2) <= 0.25:
            return 'red_flag'
        return None

#-- Tabular Editor Definition --------------------------------------------


elem_tabular_editor = TabularEditor(
    adapter=ElemTabularAdapter(),
    selected_row='current_row',
)

#-- CellArrayView Class Definition ---------------------------------------


class CellArray(HasTraits):

    data = Array

    cell_view = Instance(ICellView)

    def _cell_view_default(self):
        return CellView()

    current_row = Int(-1)

    @on_trait_change('current_row')
    def redraw(self):
        self.cell_view.redraw()

    @on_trait_change('current_row')
    def _display_current_row(self):
        if self.current_row != -1:
            self.cell_view.set_cell(self.current_row)

    view = View(
        HSplit(
            Item('data', editor=elem_tabular_editor,
                 show_label=False, style='readonly'),
            Group(Item('cell_view@', show_label=False),
                  )),
        title='Cell Array View',
        width=0.6,
        height=0.4,
        resizable=True,
        buttons=[CancelButton]
    )


# Run the demo (if invoked from the command line):
if __name__ == '__main__':

    from traits.api import Button

    class Container(HasTraits):
        show_array = Button

        def _show_array_fired(self):
            # Create the demo:
            demo = CellArray(data=random((100000, 3)))
            demo.configure_traits()
        view = View('show_array')

    c = Container()
    c.configure_traits()



'''
Created on Jan 9, 2019

@author: rch
'''
import time


def run_rerun_test(s):
    # Start calculation in a thread
    print('\nRUN the calculation thread from t = 0.0')
    s.run_thread()
    print('\nWAIT in main thread for 3 secs')
    time.sleep(3)
    print('\nPAUSE the calculation thread')
    s.pause()
    print('\nPAUSED wait 1 sec')
    time.sleep(1)
    print('\nRESUME the calculation thread')
    s.run_thread()
    print('\nWAIT in the main thread for 3 secs again')
    time.sleep(3)
    print('\nSTOP the calculation thread')
    s.stop()
    print('\nRUN a new calculation thread from t = 0.0')
    s.run_thread()
    print('\nJOIN the calculation thread into main thread to end simultaneously')
    s.join_thread()
    print('END all threads')



'''
This script is used to demonstrate the machinery
of the simulation framework on the example a simple
quadratic function. 

@author: rch
'''

import time

from simulator.api import \
    TLoop, TStep
from traits.api import \
    Int, Float
import traits.has_traits

import numpy as np

from .interaction_scripts import run_rerun_test


traits.has_traits.CHECK_INTERFACES = 2


class DemoNRTLoop(TLoop):
    '''Demonstration loop running from min to max with a defined step.
    '''

    k_max = Int(30, enter_set=True, auto_set=False)

    def eval(self):
        t_n1 = self.tline.val
        t_max = self.tline.max
        dt = self.tline.step
        U_k = self.tstep.get_state()
        while t_n1 < t_max:
            print('\ttime: %g' % t_n1, end='')
            k = 0
            # run the iteration loop
            while (k < self.k_max) and not self.user_wants_abort:
                R, dR = self.tstep.get_corr_pred(U_k, t_n1)
                if np.sqrt(R * R) < 1e-5:
                    print('\titer: %g' % k)
                    break
                dU = R / dR
                U_k += dU
                k += 1
            else:  # handle unfinished iteration loop
                if k >= self.k_max:  # add step size reduction
                    # no success abort the simulation
                    self.restart = True
                    print('Warning: '
                          'convergence not reached in %g iterations', k)
                return
            # accept the time step
            self.tstep.update_state(U_k, t_n1)
            self.hist.record_timestep(t_n1, U_k, R)
            # tline launches notifiers to announce a new step to subscribers
            self.tline.val = t_n1
            # set a new target time
            t_n1 += dt
        return


class DemoQuadFNModel(TStep):

    tloop_type = DemoNRTLoop

    R0 = Float(1.0, auto_set=False, enter_set=True)
    '''Target value of a function (load).
    '''
    U_n = Float(0.0, auto_set=False, enter_set=True)
    '''Current value of the primary variable.
    '''

    t_n = Float(0.0, auto_set=False, enter_set=True)
    '''Current value of the control variable.
    '''

    def get_corr_pred(self, U_k, t_n1):
        '''Return the value and the derivative of a function
        '''
        R = U_k * U_k - (self.R0 * t_n1)
        dR = 2 * U_k
        # handle the case of zero derivative - return small number
        dR = max(np.fabs(dR), 1e-3)
        # sleep a bit to make the calculation slower - sleep for 100 msecs
        # to demonstrate the pause-resume, stop-restart functionality
        time.sleep(0.01)
        return R, dR

    def init_state(self):
        '''Initialize state.
        '''
        self.U_n = 0.0
        self.t_n = 0.0

    def get_state(self):
        '''Get the current control and primary variables.
        '''
        return self.U_n

    def update_state(self, U_k, t_n):
        '''Update the control, primary and state variables..
        '''
        self.t_n = t_n
        self.U_n = U_k

    def record_state(self):
        '''Provide the current state for history recording.
        '''
        pass


# Construct a Simulator
model = DemoQuadFNModel()
sim = model.sim
print(sim.tloop)
run_rerun_test(sim)



'''
Created on Feb 25, 2010

@author: jakub
'''
from math import pi as Pi, cos, sin, exp, sqrt as scalar_sqrt

from ibvpy.api import RTrace, RTDofGraph, RTraceArraySnapshot
from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from ibvpy.tmodel.mats3D.mats3D_eval import MATS3DEval
from ibvpy.tmodel.mats_eval import IMATSEval
from numpy import \
     array, ones, zeros, outer, inner, transpose, dot, frompyfunc, \
     fabs, sqrt, linspace, vdot, identity, tensordot, \
     sin as nsin, meshgrid, float_, ix_, \
     vstack, hstack, sqrt as arr_sqrt, sign
from scipy.linalg import eig, inv
from traits.api import \
     Array, Bool, Callable, Enum, Float, HasTraits, \
     Instance, Int, Trait, Range, HasTraits, on_trait_change, Event, \
     Dict, Property, cached_property, Delegate
from traitsui.api import \
     Item, View, HSplit, VSplit, VGroup, Group, Spring


# from dacwt import DAC
#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS2D5PlasticBond(MATS2DEval):
    '''
    Elastic Model.
    '''

    # implements(IMATSEval)

    #---------------------------------------------------------------------------
    # Parameters of the numerical algorithm (integration)
    #---------------------------------------------------------------------------

    stress_state = Enum("plane_stress", "plane_strain")

    #---------------------------------------------------------------------------
    # Material parameters 
    #---------------------------------------------------------------------------

    E_m = Float(1.,  # 34e+3,
                 label="E_m",
                 desc="Young's Modulus",
                 auto_set=False)
    nu_m = Float(0.2,
                 label='nu_m',
                 desc="Poison's ratio",
                 auto_set=False)

    E_f = Float(1.,  # 34e+3,
                 label="E_f",
                 desc="Young's Modulus",
                 auto_set=False)
    nu_f = Float(0.2,
                 label='nu_f',
                 desc="Poison's ratio",
                 auto_set=False)

    G = Float(1.,  # 34e+3,
                 label="G",
                 desc="Shear Modulus",
                 auto_set=False)

    sigma_y = Float(.5,  # 34e+3,
                label="s_y",
                desc="Yield stress",
                auto_set=False)

    K_bar = Float(0.,  # 34e+3,
                label="K",
                desc="isotropic hardening",
                auto_set=False)

    H_bar = Float(0.,  # 34e+3,
                   label="H",
                   desc="kinematic hardening",
                   auto_set=False)

    D_el = Property(Array(float), depends_on='E_f, nu_f,E_m,nu_f,G, stress_state')

    @cached_property
    def _get_D_el(self):
        if self.stress_state == "plane_stress":
            return self._get_D_plane_stress()
        else:
            return self._get_D_plane_strain()

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #---------------------------------------------------------------------------------------------
    # View specification
    #---------------------------------------------------------------------------------------------

    view_traits = View(VSplit(Group(Item('E_m'),
                                      Item('nu_m'),
                                      Item('E_f'),
                                      Item('nu_f'),
                                      Item('G')
                                      ),
                                Group(Item('stress_state', style='custom'),
                                       Spring(resizable=True),
                                       label='Configuration parameters', show_border=True,
                                       ),
                                ),
                        resizable=True
                        )

    #-----------------------------------------------------------------------------------------------
    # Private initialization methods
    #-----------------------------------------------------------------------------------------------

    #-----------------------------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-----------------------------------------------------------------------------------------------

    def new_cntl_var(self):
        return zeros(3, float_)

    def new_resp_var(self):
        return zeros(3, float_)

    def get_state_array_size(self):
        return 3

    #-----------------------------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-----------------------------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        sigma = dot(self.D_el[:], eps_app_eng)

        # You print the stress you just computed and the value of the apparent E
        eps_n1 = float(eps_app_eng[6])  # hack for this particular case
        G = self.G

        eps_avg = eps_n1

        if sctx.update_state_on:
            eps_n = eps_avg - float(d_eps[6])
            sctx.mats_state_array[:] = self._get_state_variables(sctx, eps_n)

        # print 'state array ', sctx.mats_state_array
        eps_p_n, q_n, alpha_n = sctx.mats_state_array
        sigma_trial = self.G * (eps_n1 - eps_p_n)
        xi_trial = sigma_trial - q_n
        f_trial = abs(xi_trial) - (self.sigma_y + self.K_bar * alpha_n)
        # f_trial = -xi_trial - ( self.sigma_y + self.K_bar * alpha_n )

        sig_n1 = zeros((1,), dtype='float_')
        D_n1 = zeros((1, 1), dtype='float_')
        if f_trial <= 1e-8:
            sig_n1[0] = sigma_trial
            D_n1[0, 0] = G
        else:
            # print 'plastic'
            d_gamma = f_trial / (self.G + self.K_bar + self.H_bar)
            sig_n1[0] = sigma_trial - d_gamma * self.G * sign(xi_trial)
            D_n1[0, 0] = (self.G * (self.K_bar + self.H_bar)) / \
                            (self.G + self.K_bar + self.H_bar)
            # print 'stress ', sig_n1[0]
        sigma[6] = sig_n1[0]
        self.D_el[6, 6] = D_n1[0, 0]
        return  sigma, self.D_el

    #---------------------------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #---------------------------------------------------------------------------------------------

    def _get_state_variables(self, sctx, eps_n):

        eps_p_n, q_n, alpha_n = sctx.mats_state_array

        # Get the characteristics of the trial step
        #
        sig_trial = self.G * (eps_n - eps_p_n)
        xi_trial = sig_trial - q_n
        f_trial = abs(xi_trial) - (self.sigma_y + self.K_bar * alpha_n)

        if f_trial > 1e-8 :

            #
            # Tha last equilibrated step was inelastic. Here the
            # corresponding state variables must be calculated once
            # again. This might be expensive for 2D and 3D models. Then,
            # some kind of caching should be considered for the state
            # variables determined during iteration. In particular, the
            # computation of d_gamma should be outsourced into a separate
            # method that can in general perform an iterative computation.
            #
            d_gamma = f_trial / (self.G + self.K_bar + self.H_bar)
            eps_p_n += d_gamma * sign(xi_trial)
            q_n += d_gamma * self.H_bar * sign(xi_trial)
            alpha_n += d_gamma

        newarr = array([eps_p_n, q_n, alpha_n], dtype='float_')

        return newarr

    def _get_D_plane_stress(self):
        E_m = self.E_m
        nu_m = self.nu_m
        E_f = self.E_f
        nu_f = self.nu_f
        G = self.G
        D_stress = zeros([8, 8])
        D_stress[0, 0] = E_m / (1.0 - nu_m * nu_m)
        D_stress[0, 1] = E_m / (1.0 - nu_m * nu_m) * nu_m
        D_stress[1, 0] = E_m / (1.0 - nu_m * nu_m) * nu_m
        D_stress[1, 1] = E_m / (1.0 - nu_m * nu_m)
        D_stress[2, 2] = E_m / (1.0 - nu_m * nu_m) * (1.0 / 2.0 - nu_m / 2.0)

        D_stress[3, 3] = E_f / (1.0 - nu_f * nu_f)
        D_stress[3, 4] = E_f / (1.0 - nu_f * nu_f) * nu_f
        D_stress[4, 3] = E_f / (1.0 - nu_f * nu_f) * nu_f
        D_stress[4, 4] = E_f / (1.0 - nu_f * nu_f)
        D_stress[5, 5] = E_f / (1.0 - nu_f * nu_f) * (1.0 / 2.0 - nu_f / 2.0)

        D_stress[6, 6] = G
        D_stress[7, 7] = G
        return D_stress

    def _get_D_plane_strain(self):
        # TODO: adapt to use arbitrary 2d model following the 1d5 bond
        E_m = self.E_m
        nu_m = self.nu_m
        E_f = self.E_f
        nu_f = self.nu_f
        G = self.G
        D_strain = zeros([8, 8])
        D_strain[0, 0] = E_m * (1.0 - nu_m) / (1.0 + nu_m) / (1.0 - 2.0 * nu_m)
        D_strain[0, 1] = E_m / (1.0 + nu_m) / (1.0 - 2.0 * nu_m) * nu_m
        D_strain[1, 0] = E_m / (1.0 + nu_m) / (1.0 - 2.0 * nu_m) * nu_m
        D_strain[1, 1] = E_m * (1.0 - nu_m) / (1.0 + nu_m) / (1.0 - 2.0 * nu_m)
        D_strain[2, 2] = E_m * (1.0 - nu_m) / (1.0 + nu_m) / (2.0 - 2.0 * nu_m)

        D_strain[3, 3] = E_f * (1.0 - nu_f) / (1.0 + nu_f) / (1.0 - 2.0 * nu_f)
        D_strain[3, 4] = E_f / (1.0 + nu_f) / (1.0 - 2.0 * nu_f) * nu_f
        D_strain[4, 3] = E_f / (1.0 + nu_f) / (1.0 - 2.0 * nu_f) * nu_f
        D_strain[4, 4] = E_f * (1.0 - nu_f) / (1.0 + nu_f) / (1.0 - 2.0 * nu_f)
        D_strain[5, 5] = E_f * (1.0 - nu_f) / (1.0 + nu_f) / (2.0 - 2.0 * nu_f)

        D_strain[6, 6] = G
        D_strain[7, 7] = G
        return D_strain

    #---------------------------------------------------------------------------------------------
    # Response trace evaluators
    #---------------------------------------------------------------------------------------------

    def get_sig_norm(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return array([ scalar_sqrt(sig_eng[0] ** 2 + sig_eng[1] ** 2) ])

    def get_eps_app_m(self, sctx, eps_app_eng):
        return self.map_eps_eng_to_mtx((eps_app_eng[:3]))

    def get_eps_app_f(self, sctx, eps_app_eng):
        return self.map_eps_eng_to_mtx((eps_app_eng[3:6]))

    def get_sig_app_m(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return self.map_sig_eng_to_mtx((sig_eng[:3]))

    def get_sig_app_f(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return self.map_sig_eng_to_mtx((sig_eng[3:6]))

    def get_sig_b(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return array([[sig_eng[6], 0.], [0., sig_eng[7]]])

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {
                'eps_app_f'  : self.get_eps_app_f,
                'eps_app_m'  : self.get_eps_app_m,
                'sig_app_f'  : self.get_sig_app_f,
                'sig_app_m'  : self.get_sig_app_m,
                'sig_norm'   : self.get_sig_norm,
                'sig_b'      : self.get_sig_b, }





from numpy import zeros, hstack, meshgrid, vstack
from scipy import sparse
from scipy.sparse.linalg import eigsh
from scipy.sparse.linalg.dsolve import linsolve
from traits.api import HasTraits, Property, cached_property, Any

import numpy as np


class COOSparseMtx(HasTraits):

    assemb = Any

    ij_map = Property(depends_on='assemb.+')

    @cached_property
    def _get_ij_map(self):
        '''
        Derive the row and column indices of individual values 
        in every element matrix.
        '''

        ij_dof_map_list = []
        # loop over the list of matrix arrays
        for sys_mtx_arr in self.assemb.get_sys_mtx_arrays():

            el_dof_map = sys_mtx_arr.dof_map_arr
            ij_dof_map = zeros((el_dof_map.shape[0],
                                2,
                                el_dof_map.shape[1] ** 2,
                                ), dtype='int_')
            for el, dof_map in enumerate(el_dof_map):
                row_dof_map, col_dof_map = meshgrid(dof_map, dof_map)
                ij_dof_map[el, ...] = vstack([row_dof_map.flatten(),
                                              col_dof_map.flatten()])
            ij_dof_map_list.append(ij_dof_map)

        return ij_dof_map_list

    x_l = Property(depends_on='el_dof_map')

    @cached_property
    def _get_x_l(self):
        '''Helper property to get an array of all row indices'''
        return hstack([ij_map[:, 0, :].flatten()
                       for ij_map in self.ij_map])

    y_l = Property(depends_on='el_dof_map')

    @cached_property
    def _get_y_l(self):
        '''Helper property to get an array of all column indices'''
        return hstack([ij_map[:, 1, :].flatten()
                       for ij_map in self.ij_map])

    data_l = Property

    def _get_data_l(self):

        return hstack([sm_arr.mtx_arr.ravel()
                       for sm_arr in self.assemb.get_sys_mtx_arrays()])

    def solve(self, rhs, check_pos_dev=False):
        '''Construct the matrix and use the solver to get 
        the solution for the supplied rhs
        pos_dev - test the positive definiteness of the matrix. 
        '''
        ij = vstack((self.x_l, self.y_l))

        # Assemble the system matrix from the flattened data and
        # sparsity map containing two rows - first one are the row
        # indices and second one are the column indices.
        mtx = sparse.coo_matrix((self.data_l, ij))
        mtx_csr = mtx.tocsr()

        pos_def = True
        if check_pos_dev:
            evals_small, evecs_small = eigsh(mtx_csr, 3, sigma=0, which='LM')
            min_eval = np.min(evals_small)
            pos_def = min_eval > 1e-10

        u_vct = linsolve.spsolve(mtx_csr, rhs)
        return u_vct, pos_def



'''
Created on 01.03.2017

@author: abaktheer
'''

from scipy.interpolate import interp1d
from traits.api import HasStrictTraits, Range, Float

import numpy as np


class TimeFunctionFactory(HasStrictTraits):

    def deliver(self):
        raise NotImplementedError('deliver method not implemented')


class Monotonic(TimeFunctionFactory):
    maximum_loading = Float(1.0)
    number_of_increments = Float(50)
    t_max = Float(1.0)

    def _get_time_function(self):
        d_levels = np.linspace(0, self.maximum_loading, 2)
        d_levels[0] = 0
        d_levels.reshape(-1, 2)[:, 0] *= 0
        d_history = d_levels.flatten()
        d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1], self.number_of_increments)
                           for i in range(len(d_levels) - 1)])

        t_arr = np.linspace(0, self.t_max, len(self.d_arr))
        return interp1d(t_arr, d_arr)


class TFCyclicIncreasedSymmetric(TimeFunctionFactory):
    number_of_cycles = Float(10.0)
    maximum_loading = Float(1.0)
    number_of_increments = Float(20)
    t_max = Float(1.0)

    def _get_time_function(self):
        d_levels = np.linspace(0, self.maximum_loading, self.number_of_cycles * 2)
        d_levels.reshape(-1, 2)[:, 0] *= -1
        d_history = d_levels.flatten()
        d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1], self.number_of_increments)
                           for i in range(len(d_levels) - 1)])
        t_arr = np.linspace(0, self.t_max, len(self.d_arr))
        return interp1d(t_arr, d_arr)


class TFCyclicIncreasedNonsymmetric(TimeFunctionFactory):
    number_of_cycles = Float(10.0)
    maximum_loading = Float(1.0)
    number_of_increments = Float(20)
    t_max = Float(1.0)

    def _get_time_function(self):
        d_levels = np.linspace(0, self.maximum_loading, self.number_of_cycles * 2)
        d_levels.reshape(-1, 2)[:, 0] *= 0
        d_history = d_levels.flatten()
        d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1], self.number_of_increments)
                           for i in range(len(d_levels) - 1)])
        t_arr = np.linspace(0, self.t_max, len(self.d_arr))
        return interp1d(t_arr, d_arr)


class TFCyclicConstantSymmetric(TimeFunctionFactory):
    number_of_cycles = Float(10.0)
    maximum_loading = Float(1.0)
    number_of_increments = Float(20)
    t_max = Float(1.0)

    def _get_time_function(self):
        d_levels = np.linspace(0, self.maximum_loading, self.number_of_cycles * 2)
        d_levels.reshape(-1, 2)[:, 0] = -self.maximum_loading
        d_levels[0] = 0
        d_levels.reshape(-1, 2)[:, 1] = self.maximum_loading
        d_history = d_levels.flatten()
        d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1], self.number_of_increments)
                           for i in range(len(d_levels) - 1)])
        t_arr = np.linspace(0, self.t_max, len(self.d_arr))
        return interp1d(t_arr, d_arr)


class TFCyclicConstantNonsymmetric(TimeFunctionFactory):
    number_of_cycles = Float(10.0)
    maximum_loading = Float(1.0)
    unloading_ratio = Range(0., 1., value=0.5)
    number_of_increments = Float(20)
    t_max = Float(1.0)

    def _get_time_function(self):
        d_1 = np.zeros(1)
        d_2 = np.linspace(0, self.maximum_loading, self.number_of_cycles * 2)
        d_2.reshape(-1, 2)[:, 0] = self.maximum_loading
        d_2.reshape(-1, 2)[:, 1] = self.maximum_loading * \
                                   self.unloading_ratio
        d_history = d_2.flatten()
        d_arr = np.hstack((d_1, d_history))
        t_arr = np.linspace(0, self.t_max, len(self.d_arr))
        return interp1d(t_arr, d_arr)




from math import sqrt as scalar_sqrt

from ibvpy.api import RTrace, RTDofGraph, RTraceArraySnapshot
from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from ibvpy.tmodel.mats_eval import IMATSEval
from numpy import \
    array, zeros, transpose, dot, frompyfunc, \
    fabs, sqrt, linspace, vdot, identity, tensordot, \
    sin as nsin, meshgrid, float_, ix_, \
    vstack, hstack, sqrt as arr_sqrt, eye
from scipy.linalg import eig, inv
from traits.api import \
    Array, Bool, Callable, Enum, Float, HasTraits, \
    Int, Trait, Range, HasTraits, on_trait_change, Event, \
    Dict, Property, cached_property, Delegate
from traitsui.api import \
    Item, View, HSplit, VSplit, VGroup, Group, Spring


# Chaco imports
# from dacwt import DAC
#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS2DConduction(MATS2DEval):
    '''
    Elastic Model.
    '''

    # implements(IMATSEval)

    #-------------------------------------------------------------------------
    # Parameters of the numerical algorithm (integration)
    #-------------------------------------------------------------------------

    stress_state = Enum("plane_stress", "plane_strain", "rotational_symetry")

    #-------------------------------------------------------------------------
    # Material parameters
    #-------------------------------------------------------------------------

    k = Float(1.,
              label="k",
              desc="conduction",
              auto_set=False)

    D_mtx = Property(Array, depends_on='k')

    @cached_property
    def _get_D_mtx(self):
        return self.k * eye(2)

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #-------------------------------------------------------------------------
    # View specification
    #-------------------------------------------------------------------------

    view_traits = View(VSplit(Item('k'),
                              ),
                       resizable=True
                       )

    #-------------------------------------------------------------------------
    # Private initialization methods
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------

    def new_cntl_var(self):
        return zeros(3, float_)

    def new_resp_var(self):
        return zeros(3, float_)

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''

        # You print the stress you just computed and the value of the apparent
        # E

        return dot(self.D_mtx, eps_app_eng), self.D_mtx

    #-------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Response trace evaluators
    #-------------------------------------------------------------------------

    def get_sig_norm(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return array([scalar_sqrt(sig_eng[0] ** 2 + sig_eng[1] ** 2)])

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app': self.get_sig_app,
                'eps_app': self.get_eps_app,
                'sig_norm': self.get_sig_norm,
                'strain_energy': self.get_strain_energy}


if __name__ == '__main__':
    #-------------------------------------------------------------------------
    # Example using the mats2d_explore
    #-------------------------------------------------------------------------
    from ibvpy.tmodel.mats_explore import MATSExplore, MATS2DExplore
    mats_eval = MATS2DConduction()
    mats_explore = MATSExplore(dim=MATS2DExplore(mats_eval=mats_eval))
    mats_explore.tloop.eval()
    # mme.configure_traits( view = 'traits_view_begehung' )
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp(ibv_resource=mats_explore)
    ibvpy_app.main()



'''This example uses two dimensional grid as xmodel
with an additional geometric transformation geo_
The transformation automatically expands the strain  
tensor into 3D and augments the lateral strain 
component to zero, rendering planar strain state.

Thus, the material model gets automatically 3D 
while 2D equilibrium is considered at the global level.

The Desmorat model is applied.

Clarify the dependency between the geo_transform
and dimensionality - geo_transform does not necessarily
require the expansion to 3D plane strain.

Next step: provide the coupling of dofs 
'''

import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS2D4Q
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from ibvpy.tmodel.viz3d_tensor_field import \
    Vis3DTensorField, Viz3DTensorField
from mayavi import mlab
from simulator.api import \
    TStepBC
from simulator.xdomain.xdomain_transform import XDomainFEGridTransform

import numpy as np


def geo_trans(points):
    '''Transform a grid geometry to alayered, 
    sinusoidal discretization.
    '''
    x, y = points.T
    return np.c_[x, 50 * np.sin(x / 150 * np.pi) + y]


xdomain = XDomainFEGridTransform(coord_max=(150, 50),
                                 shape=(15, 5),
                                 geo_transform=geo_trans,
                                 fets=FETS2D4Q())

left_y = BCSlice(slice=xdomain.mesh[0, 0, 0, 0],
                 var='u', dims=[1], value=0)
left_x = BCSlice(slice=xdomain.mesh[0, :, 0, :],
                 var='u', dims=[0], value=-0.4)
right_x = BCSlice(slice=xdomain.mesh[-1, :, -1, :],
                  var='u', dims=[0], value=0.0)

mats = MATS3DDesmorat()
m = TStepBC(
    domains=[(xdomain, mats)],
    bc=[left_x, right_x, left_y],
    record={
        'strain': Vis3DTensorField(var='eps_ab'),
        #        'damage': Vis3DStateField(var='omega_a'),
        #        'kinematic hardening': Vis3DStateField(var='z_a')
    }
)
s = m.sim
s.tloop.verbose = True
s.tloop.k_max = 1000
s.tline.step = 0.05
s.run()
time.sleep(2)

strain_viz = Viz3DTensorField(vis3d=m.hist['strain'])
strain_viz.setup()
strain_viz.plot(0.0)

# damage_viz = Viz3DScalarField(vis3d=s.hist['damage'])
# damage_viz.setup()
# damage_viz.plot(0.0)
mlab.show()




from simulator.api import \
    TStep
from simulator.tloop_implicit import TLoopImplicit
from traits.api import \
    Float
import traits.has_traits

import numpy as np

from .interaction_scripts import run_rerun_test


traits.has_traits.CHECK_INTERFACES = 2


class ModelWithState(TStep):
    '''Model with a state management distinguishing .
    '''
    tloop_type = TLoopImplicit

    U_var_shape = (1,)
    '''Shape of the primary variable required by the TStepState.
    '''
    state_var_shapes = dict(
        eps_p_n=(1,),
        q_n=(1,),
        alpha_n=(1,)
    )
    '''Shape of the state variables.
    '''

    E = Float(1.,
              label="E",
              desc="Young's Modulus",
              enter_set=True,
              auto_set=False)

    sigma_y = Float(1.,
                    label="sigma_y",
                    desc="Yield stress",
                    enter_set=True,
                    auto_set=False)

    K_bar = Float(0.1,
                  label="K",
                  desc="Plasticity modulus",
                  enter_set=True,
                  auto_set=False)

    H_bar = Float(0.1,
                  label="H",
                  desc="Hardening modulus",
                  enter_set=True,
                  auto_set=False)

    def get_corr_pred(self, U_k, t_n1, eps_p_n, q_n, alpha_n):
        '''Return the value and the derivative of a function
        '''
        eps_n1 = U_k
        E = self.E

        sig_trial = self.E * (eps_n1 - eps_p_n)
        xi_trial = sig_trial - q_n
        f_trial = abs(xi_trial) - (self.sigma_y + self.K_bar * alpha_n)

        D_shape = sig_trial.shape[:-2] + (1, 1)
        D_k = np.zeros(D_shape, dtype='float_')

        sig_k = sig_trial
        D_k[...] = E
        I = np.where(f_trial > 1e-8)
        d_gamma = f_trial[I] / (self.E + self.K_bar + self.H_bar)
        sig_k[I] -= self.E * d_gamma * np.sign(xi_trial[I])
        D_k[I] = (
            (self.E * (self.K_bar + self.H_bar)) /
            (self.E + self.K_bar + self.H_bar)
        )
        return sig_k, D_k

    def update_state(self, U_k, t_n1, eps_p_n, q_n, alpha_n):
        '''In-place update of state variables. 
        '''
        print('CALLED')
        eps_n = U_k
        sig_trial = self.E * (eps_n - eps_p_n)
        xi_trial = sig_trial - q_n
        f_trial = abs(xi_trial) - (self.sigma_y + self.K_bar * alpha_n)
        I = np.where(f_trial > 1e-8)
        d_gamma = f_trial[I] / (self.E + self.K_bar + self.H_bar)
        eps_p_n[I] += d_gamma * np.sign(xi_trial[I])
        q_n[I] += d_gamma * self.H_bar * np.sign(xi_trial[I])
        alpha_n[I] += d_gamma


# Construct a Simulator
m = ModelWithState(sigma_y=0.5)
s = m.sim
run_rerun_test(s)



'''
Created on Mar 29, 2009

@author: jakub
'''

from math import pi as Pi, cos, sin, exp, sqrt as scalar_sqrt

from ibvpy.api import RTrace, RTDofGraph, RTraceArraySnapshot
from ibvpy.core.tstepper import \
    TStepper as TS
from ibvpy.tmodel.mats_eval import IMATSEval, MATSEval
from numpy import \
    array, ones, zeros, outer, inner, transpose, dot, frompyfunc, \
    fabs, sqrt, linspace, vdot, identity, tensordot, \
    sin as nsin, meshgrid, float_, ix_, \
    vstack, hstack, sqrt as arr_sqrt
from scipy.linalg import eig, inv
from traits.api import \
    Array, Bool, Callable, Enum, Float, HasTraits, \
    Instance, Int, Trait, Range, HasTraits, on_trait_change, Event, \
    Dict, Property, cached_property, Delegate
from traitsui.api import \
    Item, View, HSplit, VSplit, VGroup, Group, Spring


from .mats1D5bond import MATS1D5Bond


# Chaco imports
# from dacwt import DAC
#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS1D5BondAC(MATS1D5Bond):
    '''
    Adhesive Cohesive Bond Model
    '''

    s_cr = Float(1.,  # 34e+3,
                 label="s_cr",
                 desc="Critical Slip",
                 auto_set=False)
    tau_max = Float(1.,  # 34e+3,
                    label="T_max",
                    desc="maximal shear stress",
                    auto_set=False)
    tau_fr = Float(1.,  # 34e+3,
                   label="T_fr",
                   desc="Frictional shear stress",
                   auto_set=False)

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #-------------------------------------------------------------------------
    # View specification
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Private initialization methods
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------

    def setup(self, sctx):
        '''
        Intialize state variables.
        '''
        sctx.mats_state_array[:] = 0

    def get_state_array_size(self):
        '''
        Return number of number to be stored in state array
        @param sctx:spatial context
        '''
        return 2  # TODO: works just with linear element

    def new_cntl_var(self):
        return zeros(4, float_)  # TODO: adapt for 4-5..

    def new_resp_var(self):
        return zeros(4, float_)  # TODO: adapt for 4-5..

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        if sctx.update_state_on:
            omega_l = omega_r = 0
            if eps_app_eng[0] > self.s_cr:
                omega_l = 1
            if eps_app_eng[1] > self.s_cr:
                omega_l = 1
            sctx.mats_state_array[0] = omega_l
            sctx.mats_state_array[1] = omega_r

        omega_l = sctx.mats_state_array[0]
        omega_r = sctx.mats_state_array[1]

        if eps_app_eng.shape[0] == 4:
            D_mtx = zeros((4, 4))
            sigma = zeros(4)
        else:
            print("MATS1D5Bond: Unsupported number of strain components")

        if omega_l or eps_app_eng[0] > self.s_cr:
            D_mtx[0, 0] = 0
            sigma[0] = self.tau_fr  # tau_l
        else:
            D_mtx[0, 0] = self.tau_max / self.s_cr
            sigma[0] = D_mtx[0, 0] * eps_app_eng[0]

        if omega_r or eps_app_eng[1] > self.s_cr:
            D_mtx[1, 1] = 0.
            sigma[1] = self.tau_fr  # tau_r
        else:
            D_mtx[1, 1] = self.tau_max / self.s_cr
            sigma[1] = D_mtx[1, 1] * eps_app_eng[1]

        D_mtx[-2, -2] = self.Ef * self.Af
        D_mtx[-1, -1] = self.Em * self.Am

        sigma[-2] = self.Ef * self.Af * eps_app_eng[-2]  # sig_f
        sigma[-1] = self.Em * self.Am * eps_app_eng[-1]  # sig_m

        # You print the stress you just computed and the value of the apparent
        # E

        return sigma, D_mtx

    #-------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Response trace evaluators
    #-------------------------------------------------------------------------

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app_t1d': self.get_sig_app,
                'eps_app_t1d': self.get_eps_app,
                'shear_s': self.get_shear,
                'slip_s': self.get_slip,
                'msig_pos': self.get_msig_pos,
                'msig_pm': self.get_msig_pm}



# Demonstrate the derivation of the orthonormal basis
# of a general discretization

from ibvpy.fets import FETS2D4Q
from ibvpy.mesh.fe_grid import FEGrid
from mathkit import EPS
import mayavi.mlab as m
import numpy as np


def geo_trans(points):
    '''Transform a grid geometry to alayered, 
    sinusoidal discretization.
    '''
    x, y = points.T
    return np.c_[x, np.sin(x) + y]


fets = FETS2D4Q()
mesh = FEGrid(coord_max=(6, 4),
              shape=(20, 2),
              fets_eval=fets,
              geo_transform=geo_trans
              )

To3D = fets.vtk_expand_operator

x_Ia = mesh.X_Id
X_Ia = np.einsum('...i,...ij->...j', x_Ia, To3D)
I_Ei = mesh.I_Ei
x_Eia = X_Ia[I_Ei, :]
J_Emar = np.einsum(
    'imr,...ia->...mar', fets.dN_imr, x_Eia
)
m_0_Ema = J_Emar[..., 0]
m_2_Ema = np.einsum(
    '...i,...j,ijk->...k',
    m_0_Ema, J_Emar[..., 1], EPS
)
m_1_Ema = np.einsum(
    '...i,...j,ijk->...k',
    m_2_Ema, m_0_Ema, EPS)
M_rEma = np.array([m_0_Ema, m_1_Ema, m_2_Ema])
M_Emra = np.einsum('rEma->Emra', M_rEma)
norm_M_Emra = np.sqrt(
    np.einsum('...ij,...ij->...i', M_Emra, M_Emra)
)[..., np.newaxis]

T_Emra = M_Emra / norm_M_Emra
x_Ema = np.einsum(
    'im,...ia->...ma', fets.N_im, x_Eia
)

m.figure(bgcolor=(1, 1, 1))
colors = [(.8, .8, .8), (.8, .8, .8), (.8, .8, .8)]
for r, c in zip((0, 1, 2), colors):
    vec = m.quiver3d(x_Ema[..., 0], x_Ema[..., 1], x_Ema[..., 2],
                     T_Emra[..., r, 0], T_Emra[..., r, 1], T_Emra[..., r, 2],
                     color=c)
    vec.glyph.glyph.scaling = True
    vec.glyph.glyph.clamping = False
    vec.glyph.glyph.scale_mode = 'scale_by_vector'

# consider a constant displacement field and transform it into
# the coordinates - visualize
u_Ema = .5 * np.ones_like(x_Ema)
u_Ema[..., (1, 2)] = 0

u_Emr = np.einsum(
    '...ra,...a->...r',
    T_Emra, u_Ema
)

uT_Emra = np.einsum(
    '...r,...ra->...ra',
    u_Emr, T_Emra
)
colors = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
for r, c in zip((0, 1, 2), colors):
    vec = m.quiver3d(
        x_Ema[..., 0], x_Ema[..., 1], x_Ema[..., 2],
        uT_Emra[..., r, 0], uT_Emra[..., r, 1], uT_Emra[..., r, 2],
        color=c, scale_mode='vector')
    vec.glyph.glyph.scaling = True
    vec.glyph.glyph.clamping = False
    vec.glyph.glyph.scale_mode = 'scale_by_vector'
m.show()
